"""Data validation module"""
from paradict.typeref import TypeRef
from paradict.errors import Error, ValidationError


__all__ = ["VALID_DATATYPES", "is_valid", "validate", "Spec", "Validator"]


VALID_DATATYPES = ("dict", "list", "set", "obj", "bin",
                   "bool", "complex", "date", "datetime",
                   "float", "grid", "int", "str", "time")


def is_valid(data, schema, type_ref=None):
    """This function returns True if the given data
    successfully validates against the given schema

    [param]
    - data: some Python object (like a dict, a list, ...) that is
    part or include datatypes defined in VALID_DATATYPES.
    - schema: a valid schema. It might be a collection containing
    Spec instances and/or type-strings. The benefit of using Spec is
    that you can add a checker function that will serve as an extra
    programmatic validation.
    - type_ref: optional TypeRef object

    [return]
    Returns True or False"""
    validator = Validator(schema, type_ref=type_ref)
    try:
        validator.validate(data)
    except ValidationError as e:
        return False
    else:
        return True


def validate(data, schema, type_ref=None):
    """This function validate some data against a schema.
    Might raise a ValidationError.

    [param]
    - data: some Python object (like a dict, a list, ...) that is
    part or include datatypes defined in VALID_DATATYPES.
    - schema: a valid schema. It might be a collection containing
    Spec instances and/or type-strings. The benefit of using Spec is
    that you can add a checker function that will serve as an extra
    programmatic validation.
    - type_ref: optional TypeRef object

    [except]
    - ValidationError: Raised when an issue is encountered while validating the data
    """
    validator = Validator(schema, type_ref=type_ref)
    validator.validate(data)


class Spec:
    """A Spec can be used to form a schema along with string-types.
    The particularity of a Spec is that it can carry a checker that
    is a function to serve as an extra programmatic validation"""
    def __init__(self, datatype, checker=None):
        """
        Init
        [param]
        - datatype: a string representing a valid datatype.
        Check the VALID_DATATYPES variable to discover valid types.
        - checker: an optional function that will be called with passed
        as argument, the specific data it should check. This function
        should return a boolean to validate this piece of data

        [except]
        - ValidationError: raised if the datatype isn't a valid one
        """
        if datatype not in VALID_DATATYPES:
            msg = "Only these datatypes are valid: {}"
            msg = msg.format(" ".join(VALID_DATATYPES))
            raise ValidationError(msg)
        self._datatype = datatype
        self._checker = checker

    @property
    def datatype(self):
        return self._datatype

    @property
    def checker(self):
        return self._checker


class Validator:
    """Class to validate data against a schema"""
    def __init__(self, schema, type_ref=None):
        """
        Init

        [param]
        - schema: the schema
        - type_ref: optional TypeRef instance
        """
        self._schema = schema
        self._type_ref = type_ref if type_ref else TypeRef()


    @property
    def schema(self):
        return self._schema

    @property
    def type_ref(self):
        return self._type_ref

    def validate(self, data):
        """Validate data. Might raise a validation error"""
        if not self._validate(data, self._schema):
            raise ValidationError

    def _validate(self, target, schema):
        if target is None:
            return True
        target = self._type_ref.adapt(target)
        t = type(schema)
        if t in self._type_ref.str_types:
            return self._validate_datatype(target, schema)
        if t is Spec:
            return self._ensure_spec(target, schema)
        if t in self._type_ref.list_types:
            return self._validate_list(target, schema)
        if t in self._type_ref.set_types:
            return self._validate_set(target, schema)
        if t in self._type_ref.dict_types:
            return self._validate_dict(target, schema)
        if t in self._type_ref.obj_types:
            return self._validate_obj(target, schema)
        msg = "Invalid schema"
        raise ValidationError(msg)

    def _validate_list(self, target, schema):
        """Schema SHOULD be a list"""
        if type(target) not in self._type_ref.list_types:
            return False
        if not schema:
            return True
        for item in target:
            matched = False
            for expected_type in schema:
                if self._validate(item, expected_type):
                    matched = True
                    break
            if not matched:
                return False
        return True

    def _validate_set(self, target, schema):
        """Schema SHOULD be a set"""
        if type(target) not in self._type_ref.set_types:
            return False
        if not schema:
            return True
        for item in target:
            matched = False
            for expected_type in schema:
                if self._validate(item, expected_type):
                    matched = True
                    break
            if not matched:
                return False
        return True

    def _validate_dict(self, target, schema):
        if type(target) not in self._type_ref.dict_types:
            return False
        if not schema:
            return True
        for key, val in target.items():
            try:
                r = self._validate(val, schema[key])
            except KeyError as e:
                return False
            else:
                if not r:
                    return False
        return True

    def _validate_obj(self, target, schema):
        if type(target) not in self._type_ref.obj_types:
            return False
        if not schema:
            return True
        for key, val in target.items():
            try:
                r = self._validate(val, schema[key])
            except KeyError as e:
                return False
            else:
                if not r:
                    return False
        return True

    def _validate_datatype(self, target, datatype):
        valid_types_map = {"dict": self._type_ref.dict_types,
                           "list": self._type_ref.list_types,
                           "set": self._type_ref.set_types,
                           "obj": self._type_ref.obj_types,
                           "bin": self._type_ref.bin_types,
                           "bool": self._type_ref.bool_types,
                           "complex": self._type_ref.complex_types,
                           "date": self._type_ref.date_types,
                           "datetime": self._type_ref.datetime_types,
                           "float": self._type_ref.float_types,
                           "grid": self._type_ref.grid_types,
                           "int": self._type_ref.int_types,
                           "str": self._type_ref.str_types,
                           "time": self._type_ref.time_types}
        try:
            valid_types = valid_types_map[datatype]
        except KeyError as e:
            msg = "Unknown data type '{}'".format(datatype)
            raise ValidationError(msg)
        for valid_type in valid_types:
            if type(target) == valid_type:
                return True
        return False

    def _ensure_spec(self, target, spec):
        if not self._validate_datatype(target, spec.datatype):
            return False
        if not spec.checker:
            return True
        try:
            return bool(spec.checker(target))
        except Exception as e:
            msg = "Error while running a checker"
            raise ValidationError(msg)



"""High-level functions to deserialize Paradict binary/text data into a Python dict"""
from paradict.deserializer.decoder import Decoder
from paradict.deserializer.unpacker import Unpacker


__all__ = ["decode", "unpack"]


def decode(text, type_ref=None, receiver=None, obj_builder=None,
           root_dir=None):
    """
    Convert some textual Paradict data into a Python dictionary

    [param]
    - text: string to convert into a Python dict
    - type_ref: optional TypeRef object
    - receiver: callback function that will be called at the end of conversion.
    This callback function accepts the Decoder instance as argument
    - obj_builder: function that accepts a paradict.box.Obj container and
    returns a fresh new Python object
    - root_dir: root directory in which the attachments dir is supposed to be

    [return]
    Return the newly built Python object
    """
    decoder = Decoder(type_ref=type_ref, receiver=receiver,
                      obj_builder=obj_builder,
                      root_dir=root_dir)
    decoder.feed(text)
    if not decoder.queue.is_empty():
        decoder.feed("\n")
    decoder.feed("===\n")  # it's important to send an explicit end of stream
    return decoder.data


def unpack(raw, type_ref=None, receiver=None, obj_builder=None,
           dict_only=False):
    """
    Convert some binary Paradict data into a Python dictionary

    [param]
    - raw: raw data previously packed with Paradict
    - type_ref: optional TypeRef object
    - receiver: callback function that will be called at the end of conversion.
    This callback function accepts the Decoder instance as argument
    - obj_builder: function that accepts a paradict.box.Obj container and
    returns a fresh new Python object
    - dict_only: boolean to enforce dict as root

    [return]
    Return the newly built Python object
    """
    unpacker = Unpacker(type_ref=type_ref,
                        receiver=receiver,
                        obj_builder=obj_builder,
                        dict_only=dict_only)
    unpacker.feed(raw)
    return unpacker.data



"""This module exposes the split function that will parse a valid key-value string"""
import re
from paradict import errors, const
from collections import namedtuple


__all__ = ["Info", "split"]


Info = namedtuple("Info", ["key", "val", "sep", "mode"])


def split(val):
    """
    Split a non-empty string into key val.
    The string should follow one of these format:
    - data_mode format: key: value
    - config_mode format: key = value

    [param]
    - val: non-empty string

    [return]
    Return an Info namedtuple made of: key, val, sep, and mode attributes.
    The key and val are strings. The sep is either ":" or "=".
    The mode is either paradict.const.DATA_MODE or paradict.const.CONFIG_MODE.
    Note that if the sep is a colon, it means that the mode is DATA_MODE.
    """
    if not val or not isinstance(val, str):
        msg = "Only non-empty strings can be split"
        raise errors.Error(msg)
    r = _parse(val)
    if not r or len(r) != 3:
        raise errors.Error("Parsing error")
    left, sep, right = r
    key = left.strip().replace(r"\'", "'").replace(r'\"', '"')
    val = right.strip()
    if sep == "=":
        mode = const.CONFIG_MODE
    elif sep == ":":
        mode = const.DATA_MODE
    else:
        raise errors.Error("Missing separator character")
    return Info(key, val, sep, mode)


def _parse(val):
    """regex parse"""
    match_obj = re.fullmatch(KEY_VAL_PATTERN, val)
    groups_list = list()
    if not match_obj:
        return groups_list
    for group in match_obj.groups():
        if group is not None:
            groups_list.append(group)
    return groups_list


# double-quoted key pattern (left side of colon sign)
KEY_PATTERN_1 = r'''^(?<!\\)("[^"](?:[^"]*?(?:\\")*?)+(?<!\\)")\s*'''

# single-quoted key pattern (left side of colon sign)
KEY_PATTERN_2 = r"""^(?<!\\)('[^'](?:[^']*?(?:\\')*?)+(?<!\\)')\s*"""

# unquoted key pattern (left side of colon or equal sign)
KEY_PATTERN_3 = r"""^(.+?)\s*"""

# key pattern for config mode
KEY_PATTERN_4 = r"""^([a-zA-Z0-9_-]*)\s*"""

# value pattern (right side of colon or equal sign)
VALUE_PATTERN = r"""(.*)$"""


# Mode-based key-val patterns
DATA_MODE_PATTERN_1 = KEY_PATTERN_1 + "(:)" + VALUE_PATTERN
DATA_MODE_PATTERN_2 = KEY_PATTERN_2 + "(:)" + VALUE_PATTERN
DATA_MODE_PATTERN_3 = KEY_PATTERN_3 + "(:)" + VALUE_PATTERN
CONFIG_MODE_PATTERN = KEY_PATTERN_4 + "(=)" + VALUE_PATTERN

# Multi-mode key-val pattern
KEY_VAL_PATTERN = "|".join((CONFIG_MODE_PATTERN,
                            DATA_MODE_PATTERN_1,
                            DATA_MODE_PATTERN_2,
                            DATA_MODE_PATTERN_3))



import os

import paradict.io_text
from paradict import tags
from paradict.tags.mappings import (PINT_TO_SIZE,
                                    NINT_TO_SIZE, BIN_TO_SIZE,
                                    VARSTR_TO_SIZE, STR_TO_SIZE)


def scan(file):
    """Scan a binary Paradict file object, yielding a tag with
    the slice object of its associated payload"""
    file.seek(0)
    while 1:
        tag = file.read(1)
        if not tag:
            return
        file.seek(-1, os.SEEK_CUR)  # put cursor back to original position
        # STRx
        if tags.STR_8 <= tag <= tags.STR_256:
            slice_object = process_strx(file, tag)
        # STR
        elif tag in (tags.STR_SHORT,
                           tags.STR_MEDIUM,
                           tags.STR_LONG,
                           tags.STR_BIG,
                           tags.STR_HEAVY):
            slice_object = process_str(file, tag)
        # BIN
        elif tag in (tags.BIN_SHORT,
                           tags.BIN_MEDIUM,
                           tags.BIN_LONG,
                           tags.BIN_BIG,
                           tags.BIN_HEAVY):
            slice_object = process_bin(file, tag)
        # PINTx
        elif tags.PINT_8 <= tag <= tags.PINT_64:
            slice_object = process_pintx(file, tag)
        # PINT_BIG
        elif tag == tags.PINT_BIG:
            slice_object = process_pint_big(file, tag)
        # PINT_HEAVY
        elif tag == tags.PINT_HEAVY:
            slice_object = process_pint_heavy(file, tag)
        # NINTx
        elif tags.NINT_8 <= tag <= tags.NINT_64:
            slice_object = process_nintx(file, tag)
        # NINT_BIG
        elif tag == tags.NINT_BIG:
            slice_object = process_nint_big(file, tag)
        # NINT_HEAVY
        elif tag == tags.NINT_HEAVY:
            slice_object = process_nint_heavy(file, tag)
        # else...
        else:
            slice_object = process_tag(file, tag)
        yield tag, slice_object


def read_payload_size(file, nb):
    """
    Read the expected size of payload
    Note that nb is the number of bytes encoding the size of payload
    """
    position = file.tell()
    r = file.read(nb)
    file.seek(position)
    return int.from_bytes(r, "little", signed=False)


def get_slice_for_fixed_length_datum(file, payload_size):
    """
    This function returns the slice object for
    datums with fixed-length payload
    """
    position = file.tell()
    width = 1 + payload_size  # 1 tag + size
    start, stop = position, position + width
    file.seek(stop)
    return slice(start, stop)


def get_slice_for_variable_length_datum(file, nb):
    """
    This function is used to generate a slice object for
    tags with variable-length payload such as:
    STR_SHORT, STR_MEDIUM, STR_LONG, STR_BIG, STR_HEAVY,
    BIN_SHORT, BIN_MEDIUM, BIN_LONG, BIN_BIG, and BIN_HEAVY.

    nb is the number of bytes encoding the size of payload

    This function returns a slice object
    """
    # put the file cursor right after the tag
    position = file.tell()
    file.seek(position+1)
    # header_size
    header_size = 1 + nb  # 1 tag + nb
    payload_size = read_payload_size(file, nb) + 1  # \x00 as payload size means 1
    # Width is the size (bytes) of the datum (header + payload)
    # note that the header is made of tag + byte(s) to store payload size
    width = header_size + payload_size
    start, stop = position, position + width
    file.seek(stop)  # move the file cursor to the stop
    return slice(start, stop)


def process_tag(file, tag):
    position, width = file.tell(), 1
    start, stop = position, position + width
    file.seek(stop)  # move the file cursor to the stop
    return slice(start, stop)


def process_str(file, tag):
    # nb is the number of bytes to encode the size of payload
    nb = VARSTR_TO_SIZE[tag]
    return get_slice_for_variable_length_datum(file, nb)


def process_strx(file, tag):
    payload_size = STR_TO_SIZE[tag]
    return get_slice_for_fixed_length_datum(file, payload_size)


def process_bin(file, tag):
    # nb is the number of bytes to encode the size of payload
    nb = BIN_TO_SIZE[tag]
    return get_slice_for_variable_length_datum(file, nb)


def process_pint_big(file, tag):
    nb = 1
    return get_slice_for_variable_length_datum(file, nb)


def process_pint_heavy(file, tag):
    nb = 2
    return get_slice_for_variable_length_datum(file, nb)


def process_pintx(file, tag):
    payload_size = PINT_TO_SIZE[tag]
    return get_slice_for_fixed_length_datum(file, payload_size)


def process_nint_big(file, tag):
    return process_pint_big(file, tag)


def process_nint_heavy(file, tag):
    return process_pint_heavy(file, tag)


def process_nintx(file, tag):
    payload_size = NINT_TO_SIZE[tag]
    return get_slice_for_fixed_length_datum(file, payload_size)



"""Load and dump text Paradict from/to file"""
import os
from paradict import const
from paradict.serializer.encoder import Encoder
from paradict.deserializer.decoder import Decoder


__all__ = ["load", "dump"]


def load(file, type_ref=None, receiver=None, obj_builder=None,
         root_dir=None):
    """
    Open a textual Paradict file then read its contents into Python dict

    [param]
    - file: text file object
    - type_ref: optional TypeRef object
    - receiver: callback function that will be called at the end of conversion.
    This callback function accepts the Decoder instance as argument
    - obj_builder: function that accepts a paradict.box.Obj container and
    returns a fresh new Python object
    - root_dir: The root_dir should be set only when the file object doesn't have
        a '.name' property. The root_dir will help to load attachments.
    [return]
    Return the newly built Python object
    """
    if root_dir is None:
        try:
            root_dir = os.path.dirname(os.path.abspath(file.name))
        except AttributeError as e:
            root_dir = None
    decoder = Decoder(type_ref=type_ref, receiver=receiver,
                      obj_builder=obj_builder,
                      root_dir=root_dir)
    while True:
        line = file.readline()
        if not line:
            break
        decoder.feed(line)
    if not decoder.queue.is_empty():
        decoder.feed("\n")
    decoder.feed("===\n")
    return decoder.data


def dump(data, file, *, mode=const.DATA_MODE, type_ref=None,
          bin_to_text=False, root_dir=None, attachments_dir="attachments"):
    """
    Serialize a Python dict object with the Paradict text format then write it to a file

    [param]
    - data: Python dict object
    - file: text file object
    - mode: either const.DATA_MODE or const.CONFIG_MODE. Defaults to DATA_MODE.
    - type_ref: optional TypeRef object
    - bin_to_text: boolean to tell whether bin data should be converted into text or not
    - root_dir: the root_dir inside which attachments_dir is supposed to be.
        Set this only when bin_to_text is False and when the file object doesn't have a '.name'
        property that is basically the filename.
    - attachments_dir: path to attachments directory. Relative paths should use
        a slash as separator
    """
    if bin_to_text:
        root_dir = None
    else:
        if root_dir is None:
            try:
                root_dir = os.path.dirname(os.path.abspath(file.name))
            except AttributeError as e:
                root_dir = None
    encoder = Encoder(mode=mode, type_ref=type_ref,
                      bin_to_text=bin_to_text, root_dir=root_dir,
                      attachments_dir=attachments_dir)
    for r in encoder.encode(data):
        file.write(r+"\n")



from paradict.serializer import encode, pack
from paradict.serializer.encoder import Encoder
from paradict.serializer.packer import Packer
from paradict.deserializer import decode, unpack
from paradict.deserializer.scanner import scan
from paradict.deserializer.decoder import Decoder
from paradict.deserializer.unpacker import Unpacker
from paradict.typeref import TypeRef
from paradict.validator import is_valid, Validator
from paradict.kv import split as split_kv
from paradict.misc import forge_bin, stringify_bin
from paradict.const import CONFIG_MODE, DATA_MODE, Datatype


__all__ = ["encode", "decode",
           "pack", "unpack", "scan",
           "validate", "is_valid", "split_kv",
           "forge_bin", "stringify_bin",
           "Encoder", "Decoder",
           "Packer", "Unpacker", "Datatype",
           "TypeRef", "Validator", "CONFIG_MODE", "DATA_MODE"]



"""A FIFO queue for processing textual Paradict data"""
from collections import namedtuple


__all__ = ["BufferedTextStream"]


class BufferedTextStream:
    """A FIFO queue for processing textual Paradict data"""
    def __init__(self):
        self._buffer = list()
        self._partial_s = ""

    def is_empty(self):
        if not self._buffer and not self._partial_s:
            return True
        return False

    def put(self, s):
        """Store textual data in the buffer. This data will then be iteratively
        extracted by the 'get' method, line by line.
        Note: make sure that each line is ended with a newline '\n' character"""
        if not s:
            return
        # split lines and store first and last
        lines = s.splitlines(keepends=True)
        first = lines.pop(0)
        try:
            last = lines.pop()
        except IndexError as e:
            last = ""
        # handle self._partial_s
        if self._partial_s:
            self._buffer.append(self._partial_s + first)
            self._partial_s = ""
        else:
            self._buffer.append(first)
        # update buffer with splitted lines
        self._buffer.extend(lines)
        # update buffer/partial_s with the value of last
        if last:
            if last.endswith("\n"):
                self._buffer.append(last)
            else:
                self._partial_s = last

    def get(self):
        if self._buffer:
            return self._buffer.pop(0)

    def get_all(self):
        """Generator for iteratively getting each line composing the
        textual data stored in the buffer.
        Note that lines yielded won't end with a newline '\n' character"""
        while self._buffer:
            yield self._buffer.pop(0)
